#include "do_game.h"
#include "globals.h"

#include <stdlib.h>
#include <stdio.h>

#include "hw_md.h"
#include "gfx.h"

#include "font.h"
#include "border.h"
#include "block.h"
#include "backgrounds.h"

#include "tetris.h"

#include "program_status.h"

uint8_t do_game() {
	prev_buttons |= SEGA_CTRL_START;

	uint32_t more = 1;

	tile_rectangle(0x0000, PLANE_A, 0, 0, 40, 28);
	tile_rectangle(0x0000, PLANE_B, 0, 0, 40, 28);

	set_palette(font_palette_game, 0x00, sizeof(font_palette_game)/sizeof(uint16_t));
	SET_VDP_REG(7, 0x10); // Backdrop color is the first background palette entry.
	set_palette(background_palettes[0], 0x10, 16);
	set_palette(border_palette, 0x20, sizeof(border_palette)/sizeof(uint16_t));
	set_palette(block_palette, 0x30, sizeof(block_palette)/sizeof(uint16_t));

	SET_VDP_REG(15, 2);

	// Tile 0 must always remain all zeros for its pattern, and use palette
	// 0 entry 0 thus being transparent, so that plane A has a transparent
	// tile to use to let the plane B background show through.

	// Tile 1 is an opaque black tile, used inside the playfield for empty
	// blocks. It reuses the border palette's black color.
	ACCESS_VDP(VRAM_W, TILE_LOC(1));
	for (size_t i = 0; i < 8; i++) {
		DATA_PORT_LONG_WORD(0x22222222);
	}

	ACCESS_VDP(VRAM_W, TILE_LOC(' ') + sizeof(font_patterns));
	for (size_t i = 0; i < sizeof(border_patterns)/sizeof(uint32_t); i++) {
		DATA_PORT_LONG_WORD(border_patterns[i]);
	}

	ACCESS_VDP(VRAM_W, TILE_LOC(0x10));
	for (size_t i = 0; i < sizeof(block_patterns)/sizeof(uint32_t); i++) {
		DATA_PORT_LONG_WORD(block_patterns[i]);
	}

	const uint16_t background_start = TILE_LOC(' ') + sizeof(font_patterns) + sizeof(border_patterns);
	ACCESS_VDP(VRAM_W, background_start);
	for (size_t i = 0; i < LONGWORDS_PER_BACKGROUND; i++) {
		DATA_PORT_LONG_WORD(backgrounds[0][i]);
	}

	print_string("NEXT", PLANE_A, 14, 2);

	print_string("SCORE", PLANE_A, 7, 3);
	print_string("LINES", PLANE_A, 7, 6);
	print_string("SCORE", PLANE_A, 7, 9);

	SET_VDP_REG(15, 2);
	for (uint32_t y = 0; y < 23; y++) {
		ACCESS_VDP(VRAM_W, PLANE_A + SCREEN_LOC(14, 3 + y));
		for (uint32_t x = 0; x < 12; x++) {
			uint16_t tile = border_tiles[y*12 + x];
			if (tile == 0xFFFF) {
				DATA_PORT_WORD(0x0001 | 0x4000);
			}
			else {
				DATA_PORT_WORD((tile + ' ' + 16*4) | 0x4000);
			}
		}
	}

	for (uint16_t y = 0; y < 28; y++) {
		ACCESS_VDP(VRAM_W, PLANE_B + SCREEN_LOC(0, y));
		for (uint16_t x = 0; x < 40; x++) {
			DATA_PORT_WORD(((background_start/32) + x + y * 40) | 0x2000);
		}
	}

	tetris t;
	tetris_initialize(&t, 0);
	while (more) {
		buttons = get_pad(0);
		uint16_t input = INP_NONE;

#define ADD_INP(inp) if (buttons & SEGA_CTRL_##inp) input |= INP_##inp
		ADD_INP(UP);
		ADD_INP(DOWN);
		ADD_INP(RIGHT);
		ADD_INP(LEFT);
		ADD_INP(A);
		ADD_INP(B);
		ADD_INP(C);

		int8_t background_sequence[] = {
			0,	// Level 0
			-1,	// Level 1
			1,	// Level 2
			-1,	// Level 3
			2,	// Level 4
			-1,	// Level 5
			3,	// Level 6
			-1,	// Level 7
			4,	// Level 8
			5,	// Level 9
			6,	// Level 10
			7,	// Level 11
			-1,	// Level 12
			8,	// Level 13
			-1,	// Level 14
			9	// Level 15
		};

		uint8_t prev_level = t.level;
		uint8_t prev_next = t.next;
		tetris_do_frame(&t, input);

		if (t.level > prev_level) {
			if (t.level <= 15) {
				int8_t background_index = background_sequence[t.level];
				if (background_index > 0) {
					set_palette(background_palettes[background_index], 0x10, 16);
					ACCESS_VDP(VRAM_W, background_start);
					for (size_t i = 0; i < LONGWORDS_PER_BACKGROUND; i++) {
						DATA_PORT_LONG_WORD(backgrounds[background_index][i]);
					}
				}
			}
		}

		/* Place tiles for the field. */
		for (uint16_t y = 0; y < FIELD_HEIGHT; y++) {
			ACCESS_VDP(VRAM_W, PLANE_A + SCREEN_LOC(15, 4 + y));
			for (uint16_t x = 0; x < FIELD_WIDTH; x++) {
				if (t.field[y][x] == 7) {
					DATA_PORT_WORD(0x0001 | 0x4000);
				}
				else {
					DATA_PORT_WORD((0x0010 + t.field[y][x]) | 0x6000);
				}
			}
		}

		/* Place tiles for the next piece. */
		if (t.next < 7 && (prev_next != t.next)) {
			tile_rectangle(0x0000, PLANE_A, 18, 1, 4, 2);
			for (uint8_t i = 0; i < 4; i++) {
				ACCESS_VDP(VRAM_W, PLANE_A + SCREEN_LOC(pcs[t.next][0][0][i] + 18, pcs[t.next][0][1][i] + !t.next));
				DATA_PORT_WORD((0x0010 + t.next) | 0x6000);
			}
		}

		/* Place tiles for the active piece. */
		if (t.game_status == GAME_ACTIVE || t.game_status == GAME_OVER) {
			for (uint8_t i = 0; i < 4; i++) {
				ACCESS_VDP(VRAM_W, PLANE_A + SCREEN_LOC(pcs[t.pctype][t.pcdir][0][i] + 15 + t.pcx, pcs[t.pctype][t.pcdir][1][i] + 4 + t.pcy));
				DATA_PORT_WORD((0x0010 + t.pctype) | 0x6000);
			}
		}

		char number_string[8];

		snprintf(number_string, 8, "% 7d", (int)t.score);
		print_string(number_string, PLANE_A, 5, 4);
		snprintf(number_string, 8, "% 7d", (int)t.lines);
		print_string(number_string, PLANE_A, 5, 7);
		snprintf(number_string, 8, "% 7d", (int)t.level);
		print_string(number_string, PLANE_A, 5, 10);

		if (!(prev_buttons & SEGA_CTRL_START) && (buttons & SEGA_CTRL_START)) {
			more = 0;
		}

		wait_for_vblank();
		prev_buttons = buttons;
	}

	SET_VDP_REG(7, 0x00);
	return PROGRAM_MENU;
}
