#include "tetris.h"
#include "globals.h"

#define WAIT_DUR 120 /* How many frames the player has to wait before the first ARE begins, then active gameplay starts. (So that's a total of WAIT_DUR + ARE_DUR frames of waiting before active gameplay begins.) I've recently discovered this is actually the correct number of frames. */ 
#define AUTOSHIFT_DUR 20 /* How many frames the player must hold left/right before autoshift begins. */
#define CLEAR_DUR 42 /* How many frames there are between line clearing starting, ending, then ARE starting. */
#define ARE_DUR 30 /* ARE means "A Restoring Emptiness" for English readers; it has a different meaning in Japanese (coined by Japanese TGM players, I think), but the English backronym is approximately the intended interpretation, given the contexts its used in. It's the waiting period between the stack entering a stable state after a piece locks/lines clear and the next piece entering. An alternative name might be "piece entry delay." */
#define SLIDE_DUR 30 /* How many frames a piece can sit/slide on the bottom of the matrix or stack surface before locking. */

/* pcs[pctype][pcdir][axis][which_block]
 *
 * axis == 0 is x, axis == 1 is y.
 *
 * See tetris.h for pctype.
 *
 * pcdir goes 0 (piece entrance direction), then CCW rotations to 1, 2, 3, then
 * back again. */
const uint8_t pcs[7][4][2][4] = {
	{{{0,1,2,3},{1,1,1,1}},
	 {{2,2,2,2},{0,1,2,3}},
	 {{0,1,2,3},{1,1,1,1}},
	 {{2,2,2,2},{0,1,2,3}}},
	{{{0,1,1,2},{1,1,2,2}},
	 {{2,1,2,1},{0,1,1,2}},
	 {{0,1,1,2},{1,1,2,2}},
	 {{2,1,2,1},{0,1,1,2}}},
	{{{1,2,0,1},{1,1,2,2}},
	 {{0,0,1,1},{0,1,1,2}},
	 {{1,2,0,1},{1,1,2,2}},
	 {{0,0,1,1},{0,1,1,2}}},
	{{{1,2,1,2},{1,1,2,2}},
	 {{1,2,1,2},{1,1,2,2}},
	 {{1,2,1,2},{1,1,2,2}},
	 {{1,2,1,2},{1,1,2,2}}},
	{{{0,1,2,1},{1,1,1,2}},
	 {{1,1,2,1},{0,1,1,2}},
	 {{1,0,1,2},{1,2,2,2}},
	 {{1,0,1,1},{0,1,1,2}}},
	{{{0,1,2,0},{1,1,1,2}},
	 {{1,1,1,2},{0,1,2,2}},
	 {{2,0,1,2},{1,2,2,2}},
	 {{0,1,1,1},{0,0,1,2}}},
	{{{0,1,2,2},{1,1,1,2}},
	 {{1,2,1,1},{0,0,1,2}},
	 {{0,0,1,2},{1,2,2,2}},
	 {{1,1,0,1},{0,1,2,2}}}};

/* speed_level[difficulty][level]
 * Quantities are in frames/row. */
static const unsigned char speed_level[4][16] = {
	{48, 32, 24, 18, 14, 12, 10, 8, 6, 4, 12, 10, 8, 6, 4, 2},
	{48, 24, 18, 15, 12, 10,  8, 6, 4, 2, 10,  8, 6, 4, 2, 1},
	{40, 20, 15, 12, 10,  8,  6, 4, 2, 1, 10,  8, 6, 4, 2, 1},
	{30, 15, 12, 10,  8,  6,  4, 2, 1, 1,  8,  6, 4, 2, 1, 1}};

/* time_level[level]
 * Quantities are in 256_frames. So 1 here means 256 frames. */
static const unsigned char time_level[16] = {14, 9, 9, 9, 9, 9, 9, 9, 9, 14, 14, 9, 9, 9, 9, 14};

static const unsigned score_base[4] = {100, 400, 900, 2000};

int intersecting(tetris * tp);
int filled_row(tetris * tp, int row);

void tetris_initialize(tetris * tp, uint8_t difficulty) {
	tp->score = 0;
	tp->lines = 0;
	tp->level = 0;
	tp->pcx = 0;
	tp->pcy = 0;
	tp->pctype = 7; /* pctype == 7 indicates the piece isn't valid. */
	tp->pcdir = 0;
	for (unsigned y = 0; y < FIELD_HEIGHT; y++) {
		for (unsigned x = 0; x < FIELD_WIDTH; x++) {
			tp->field[y][x] = 7;
		}
	}
	tp->sound = 0; // it'll be a while before sound is working...
	tp->music = 0; // same for music
	// actually, I could probably just have some visual indicators showing the sounds that would play, if it were working; music too

	tp->rand_state = 0; // TODO: Save and restore rand_state from a file, to have persistence like the AC version. It's like the bootlegs right now.
	tp->difficulty = difficulty % 4; // TODO: Same with difficulty.
	// TODO: Implement the other state the AC version stored, like number
	// of games played; it's relevant to players, in that they can know the
	// state of the randomizer with that info.
	tp->wait_frames = 0;
	tp->level_frames = 0;
	tp->autoshift_frames = 0;
	tp->fall_frames = 0;
	tp->slide_frames = 0;
	tp->clear_frames = 0;
	tp->are_frames = 0;
	for (unsigned i = 0; i < 4; i++) {
		tp->line_rows[i] = -1;
	}
	tp->lines_this_level = 0;
	tp->game_status = GAME_WAIT;
	tp->prev_input = INP_NONE;
}

void tetris_do_frame(tetris * tp, uint16_t input) {
	/* In AC version, if left/right+down is pressed, it's always read as left/right.
	 * Also, left+right+down isn't read as any input.
	 * Seems just left,
	 * just right,
	 * just down,
	 * just left+down,
	 * just right+down,
	 * and any of the above with only one rotation input per frame is valid.
	 * Multiple rotation inputs per frame aren't read as an input.
	 * Actually, it seems it reads multiple rotation inputs per frame as
	 * just one, which seems fine enough. */
	int tpcx;
	int tpcy;
	uint8_t tpcdir;
	int pushing = 0;
	int i, j;
	int new_lines;

	/* Processing of left/right holding is kept track of at all times. */
	if ((tp->prev_input & INP_RIGHT) && !(tp->prev_input & INP_LEFT) && (input & INP_RIGHT) && !(input & INP_LEFT)) {
		if (tp->autoshift_frames < AUTOSHIFT_DUR) {
			tp->autoshift_frames++;
		}
	}
	else if ((tp->prev_input & INP_LEFT) && !(tp->prev_input & INP_RIGHT) && (input & INP_LEFT) && !(input & INP_RIGHT)) {
		if (tp->autoshift_frames < AUTOSHIFT_DUR) {
			tp->autoshift_frames++;
		}
	}
	else {
		tp->autoshift_frames = 0;
	}

	/* Guard against overflow. The AC tetris actually didn't do
	 * this; if, somehow, you got the counter to 65535, it'd flip
	 * back to 0. (I hacked the position in memory where this data
	 * is stored.) It might actually be impossible, given the
	 * constraints of the game rules, to get the counter that high,
	 * though. */
	if (tp->game_status >= GAME_ACTIVE && tp->level_frames < 0xFFFF) {
		tp->level_frames++;
	}

	switch (tp->game_status) {
		/* The game's over when the entering piece is forced to intersect with
		 * the stack. */
		default:
		case GAME_OVER:
			// TODO: The monkey. It's relevant to gameplay, because rand_state changes when the monkey animation is selected, affecting the sequence for the next game.
			break;
		/* The "waiting period" is the frames in which the player waits for the
		 * first piece to enter, looking at an empty field. */
		case GAME_WAIT:
			tp->wait_frames++;
			if (tp->wait_frames == 1) {
				new_seq_a(tp->seq, SEQ_LEN, &seed);
				tp->seq_pos = 0;
				tp->next = tp->seq[tp->seq_pos++];
			}
			else if (tp->wait_frames == WAIT_DUR) {
				tp->game_status = GAME_ARE;
			}
			break;
		/* The "active period" is the period in which there is an active piece
		 * yet to be locked. */
		case GAME_ACTIVE:
			/* Level up happens when you clear four lines, or when
			 * tp->level_frames/256 == time_level[tp->level] upon locking a
			 * piece and you didn't clear a line. If level_frames is the
			 * proper value upon locking a piece but you clear 1-3 lines,
			 * level up is deferred until you lock a piece clearing 0 or 4
			 * lines. So it looks like level up has to be handled somewhere
			 * around the line clearing code. */
			/* Try rotating if there's a rotation input. */
			if (
					(!(tp->prev_input & INP_A) && (input & INP_A)) ||
					(!(tp->prev_input & INP_B) && (input & INP_B)) ||
					(!(tp->prev_input & INP_C) && (input & INP_C))
			   ) {
				/* Code here will change when both rotation directions are supported. */
				tpcdir = tp->pcdir;
				tp->pcdir++;
				tp->pcdir %= 4;
				if (intersecting(tp)) {
					tp->pcdir = tpcdir;
				}
			}

			/* Now try left/right shifting if there's a left/right input or the autoshift duration has passed. */
			if (tp->autoshift_frames == 1 || tp->autoshift_frames == AUTOSHIFT_DUR) {
				tpcx = tp->pcx;
				/* When autoshift_frames > 0, only one of INP_RIGHT/INP_LEFT is guaranteed to be in input. */
				if (input & INP_RIGHT) {
					tp->pcx++;
				}
				else if (input & INP_LEFT) {
					tp->pcx--;
				}
				if (intersecting(tp)) {
					tp->pcx = tpcx;
				}
			}

			/* Now push or gravitate. */
			if (input & INP_DOWN || tp->fall_frames == speed_level[tp->difficulty][(tp->level > 15) ? 15 : tp->level]) {
				if (input & INP_DOWN) {
					pushing = 1; /* Automatically locks if dropping from 1 or 0 rows above the bottom. */
				}
				tpcy = tp->pcy;
				tp->pcy++;
				if (intersecting(tp)) {
					tp->pcy = tpcy;
				}
				else if (pushing) {
					tp->score++;
				}
			}

			/* Lastly, process slide frames and lock if needed; also clear lines after locking, if needed. */
			tpcy = tp->pcy;
			tp->pcy++;
			if (intersecting(tp)) {
				tp->pcy = tpcy;
				tp->fall_frames = 0;
				// TODO: Play active-piece-touches-stack sound here.
				if (tp->slide_frames == SLIDE_DUR || pushing) {
					/* Lock the active piece. */
					for (i = 0; i < 4; i++) {
						tp->field[pcs[tp->pctype][tp->pcdir][1][i] + tp->pcy][pcs[tp->pctype][tp->pcdir][0][i] + tp->pcx] = tp->pctype;
					}

					/* Mark any lines. (A line is a row
					 * completely filled with blocks.) The
					 * marking is for the renderer, so it
					 * can show a simple line clearing
					 * animation. */
					tp->game_status = GAME_ARE;
					for (i = 0, j = FIELD_HEIGHT-1; j >= 0; j--) {
						if (filled_row(tp, j) > 0) {
							tp->line_rows[i] = j;
							i++;
							tp->game_status = GAME_CLEAR;
						}
					}
				}
				else tp->slide_frames++;
			}
			else {
				tp->pcy = tpcy;
				if (tp->fall_frames == speed_level[tp->difficulty][(tp->level > 15) ? 15 : tp->level]) tp->fall_frames = 0;
				tp->fall_frames++;
			}
			break;
		case GAME_CLEAR:
			tp->clear_frames++;

			if (tp->clear_frames == CLEAR_DUR) {
				tp->game_status = GAME_ARE;
			}
			break;
		case GAME_ARE:
			/* Looks like the game locks your piece, visually clears lines
			 * and does the clear delay, then once clear delay is over the
			 * number of lines cleared/level are updated on the first ARE
			 * frame.  */

			tp->are_frames++;

			new_lines = 0;
			if (tp->clear_frames == CLEAR_DUR) {
				int full_row, holed_row;

				tp->clear_frames = 0;

				for (full_row = FIELD_HEIGHT-1; full_row >= 0; full_row--) {
					if (filled_row(tp, full_row) > 0) break;
				}
				for (holed_row = full_row-1; holed_row >= 0; holed_row--) {
					if (filled_row(tp, holed_row) == 0) break;
				}
				for (; holed_row >= 0; full_row--) {
					for (i = 0; i < FIELD_WIDTH; i++) {
						tp->field[full_row][i] = tp->field[holed_row][i];
					}
					for (holed_row--; holed_row >= 0; holed_row--) {
						if (filled_row(tp, holed_row) == 0) break;
					}
				}
				new_lines = full_row + 1;
				for (; full_row >= 0; full_row--) {
					for (i = 0; i < FIELD_WIDTH; i++) tp->field[full_row][i] = 7;
				}

				if (new_lines > 0) {
					int multiplier;

					if (tp->level > 7) {
						multiplier = 5;
					}
					else {
						multiplier = (tp->level / 2) + 1;
					}
					tp->score += score_base[new_lines-1] * multiplier;
					if (tp->score > 999999) tp->score = 999999;

					tp->lines += new_lines;
					if (tp->lines > 999) tp->lines = 999;
					
					tp->lines_this_level += new_lines;
					// TODO: The stack-drop-after-line-clear sound.
					// It plays even if there aren't any blocks
					// that drop, but 1 or more lines were cleared.
				}
			}

			if (tp->are_frames == 1) {
				if (tp->lines_this_level >= 4 || (new_lines == 0 && (tp->level_frames / 256) == time_level[(tp->level > 15) ? 15 : tp->level])) {
					tp->level_frames = 0;
					if (tp->level < 99) {
						tp->level++;
					}
					tp->lines_this_level = 0;
					// TODO: The level-up sound. It plays even when level-up conditions are met at level 99.
				}
			}

			if (tp->are_frames == ARE_DUR) {
				tp->are_frames = 0;
				tp->pcx = 3;
				tp->pcy = -1;
				tp->pctype = tp->next;
				tp->next = tp->seq[tp->seq_pos];
				tp->seq_pos = (tp->seq_pos + 1) % SEQ_LEN;
				tp->pcdir = 0;
				tp->slide_frames = 0;
				for (i = 0; i < 4; i++) {
					tp->line_rows[i] = -1;
				}

				if (intersecting(tp)) {
					tp->game_status = GAME_OVER;
				} else {
					tp->game_status = GAME_ACTIVE;
				}
			}
			break;
	}

	tp->prev_input = input;
}

/* Checks if the piece is intersecting with the stack or the outside of the
 * field.  */
int intersecting(tetris * tp) {
	if (tp->pctype == 7) {
		return 0;
	}

	for (unsigned i = 0; i < 4; i++) {
		int16_t tx = pcs[tp->pctype][tp->pcdir][0][i] + tp->pcx;
		int16_t ty = pcs[tp->pctype][tp->pcdir][1][i] + tp->pcy;
		if (tx < 0 || tx >= FIELD_WIDTH || ty < 0 || ty >= FIELD_HEIGHT) {
			return 1;
		}
		else if (tp->field[ty][tx] != 7) {
			return 1;
		}
	}

	return 0;
}

int filled_row(tetris * tp, int row) {
	if (row < 0 || row >= FIELD_HEIGHT) {
		return -1;
	}
	for (unsigned i = 0; i < FIELD_WIDTH; i++) {
		if (tp->field[row][i] == 7) return 0;
	}

	return 1;
}
